@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2020, 2024, Oracle and/or its affiliates.
+ * Copyright (c) 2020, 2025, Oracle and/or its affiliates.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2.0,
@@ -25,6 +25,7 @@
 
 #include "modules/util/dump/compatibility.h"
 
+#include <cassert>
 #include <regex>
 #include <unordered_map>
 #include <utility>
@@ -73,12 +74,14 @@ const std::set<std::string> k_mysqlaas_allowed_privileges = {
     "BACKUP_ADMIN",
     "CONNECTION_ADMIN",
     "FLUSH_OPTIMIZER_COSTS",
+    "FLUSH_PRIVILEGES",
     "FLUSH_STATUS",
     "FLUSH_TABLES",
     "FLUSH_USER_RESOURCES",
     "REPLICATION_APPLIER",
     "ROLE_ADMIN",
     "SHOW_ROUTINE",
+    "TRANSACTION_GTID_TAG",
     "XA_RECOVER_ADMIN",
 };
 
@@ -1920,6 +1923,17 @@ bool parse_grant_statement(const std::string &statement,
       result.level = object_is_routine ? Privilege_level_info::Level::ROUTINE
                                        : Privilege_level_info::Level::TABLE;
     }
+
+    // TO
+    it.next_token();
+  }
+
+  result.account = get_account(&it);
+
+  if (is_role) {
+    result.with_grant = it.consume_tokens("WITH", "ADMIN", "OPTION");
+  } else {
+    result.with_grant = it.consume_tokens("WITH", "GRANT", "OPTION");
   }
 
   for (auto &p : privileges) {
@@ -1931,6 +1945,27 @@ bool parse_grant_statement(const std::string &statement,
   return true;
 }
 
+std::string to_grant_statement(const Privilege_level_info &info) {
+  assert(Privilege_level_info::Level::SCHEMA == info.level);
+
+  std::string result = info.grant ? "GRANT" : "REVOKE";
+
+  result += ' ';
+  result += shcore::str_join(info.privileges, ", ");
+  result += " ON ";
+  result += shcore::quote_identifier(info.schema);
+  result += ".* ";
+  result += info.grant ? "TO" : "FROM";
+  result += ' ';
+  result += info.account;
+
+  if (info.with_grant) {
+    result += " WITH GRANT OPTION";
+  }
+
+  return result;
+}
+
 bool supports_set_any_definer_privilege(const mysqlshdk::utils::Version &v) {
   return v >= mysqlshdk::utils::Version(8, 2, 0);
 }
