@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2020, 2024, Oracle and/or its affiliates.
+ * Copyright (c) 2020, 2025, Oracle and/or its affiliates.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2.0,
@@ -55,6 +55,7 @@
 #include "mysqlshdk/libs/db/mysql/session.h"
 #include "mysqlshdk/libs/storage/ifile.h"
 #include "mysqlshdk/libs/textui/text_progress.h"
+#include "mysqlshdk/libs/utils/atomic_flag.h"
 #include "mysqlshdk/libs/utils/synchronized_queue.h"
 
 namespace mysqlsh {
@@ -78,7 +79,7 @@ class Dump_loader {
 
   void hard_interrupt();
 
-  void abort();
+  void interruption_notification();
 
   void run();
 
@@ -702,8 +703,8 @@ class Dump_loader {
   std::unordered_set<std::string> m_skip_schemas;
   std::unordered_set<std::string> m_skip_tables;
   std::vector<std::exception_ptr> m_thread_exceptions;
-  volatile bool m_worker_hard_interrupt = false;
-  bool m_worker_interrupt = false;
+  shcore::atomic_flag m_worker_hard_interrupt;
+  shcore::atomic_flag m_worker_interrupt;
   bool m_abort = false;
 
   std::string m_character_set;
@@ -733,6 +734,8 @@ class Dump_loader {
   uint64_t m_indexes_recreated;
   // (this variable does not change once DDL finishes loading)
   uint64_t m_indexes_to_recreate = 0;
+  // whether to query for partial index progress
+  bool m_query_index_progress = true;
   // these variables are used to track the progress
   std::mutex m_indexes_progress_mutex;
   uint64_t m_indexes_completed = 0;
